# Importing necessary libraries for numerical operations and ODE integration
import numpy as np
from scipy.integrate import odeint

# Define a function to model the PID controller for a system.
def modelPID(x, t, Kp, Ki, Kd):
    # Constants and parameters used in the equations
    w = 100  # assumed weight
    # Kinetic rates for insulin action
    ka1 = 0.006
    ka2 = 0.06
    ka3 = 0.03
    # Kinetic rates for insulin degradation
    kb1 = 0.0034
    kb2 = 0.056
    kb3 = 0.024
    u_b = 0.0555  # basal insulin
    tmaxI = 55  # time to maximum insulin action
    VI = 0.12 * w  # insulin volume
    ke = 0.138  # elimination rate of insulin
    k12 = 0.066  # rate of transfer from compartment 1 to 2
    VG = 0.16 * w  # glucose volume
    F01 = 0.0097 * w  # non-insulin mediated glucose uptake
    FR = 0  # renal glucose clearance
    EGP0 = 0.0161 * w  # endogenous glucose production
    AG = 0.8  # glucose effectiveness
    Gmolar = 180.1559  # molar mass of glucose
    tmaxG = 40  # time to maximum glucose action
    sp = 110 * VG / 18  # set point for glucose

    # Current state variables for the model
    Q1, Q2, S1, S2, I, x1, x2, x3, tau, Dg, Ie, u = x
    
    # Differential equations describing the system dynamics
    # Equations for glucose kinetics
    dQ1 = -F01 - x1 * Q1 + k12 * Q2 - FR + EGP0 * (1 - x3) + (Dg * AG * 1000 / Gmolar) * tau * np.exp(-tau / tmaxG) / (tmaxG ** 2)
    dQ2 = x1 * Q1 - (k12 + x2) * Q2
    # Equation for insulin effect
    dIe = sp - Q1
    de = -dQ1
    # PID controller for insulin dose
    uNew = max(Ki * Ie + Kp * dIe + Kd * de, 0)
    # Equations for insulin kinetics
    dS1 = uNew + u_b - S1 / tmaxI
    dS2 = (S1 - S2) / tmaxI
    dI = S2 / (tmaxI * VI) - ke * I
    # Equations for insulin action dynamics
    dx1 = -ka1 * x1 + kb1 * I
    dx2 = -ka2 * x2 + kb2 * I
    dx3 = -ka3 * x3 + kb3 * I
    # Time and meal bolus dynamics
    dtau = 1
    dDg = 0
    
    # Combine all differential changes into a single vector
    dxdt = [dQ1, dQ2, dS1, dS2, dI, dx1, dx2, dx3, dtau, dDg, dIe, uNew - u]

    return dxdt

# Function to simulate the system over time given meal times and glucose doses
def simulation(timeOfMeals, dGs, model):
    tstep = 1.0  # time step for simulation
    # Initial conditions for the model
    ytot = [97.77, 19.08024, 3.0525, 3.0525, 0.033551, 0.01899, 0.03128, 0.02681, 0.0, 0, 0, 0]
    x0 = ytot
    ttot = [0]  # initial time
    
    # Loop through each meal time and simulate the system
    for i in range(len(timeOfMeals)):
        x0[8] = 0  # reset tau
        x0[9] = dGs[i]  # set meal glucose dose
        # Time vector for the current meal
        time = np.arange(tstep, timeOfMeals[i], tstep)
        # Integrate the model equations over time
        y = odeint(model, x0, time)
        # Update initial conditions for the next iteration
        x0 = y[-1, :]
        # Collect the system's response
        ytot = np.vstack([ytot, y])
        ttot = np.hstack([ttot, time + ttot[-1]])
    
    return ttot, ytot
